'' examples/manual/threads/condcreate2.bas
''
'' NOTICE: This file is part of the FreeBASIC Compiler package and can't
''         be included in other distributions without authorization.
''
'' See Also: http://www.freebasic.net/wiki/wikka.php?wakka=KeyPgCondCreate
'' --------

'Visual example of mutual exclusion + mutual synchronization between 2 threads
'by using Mutex and CondVar:
'the "user-defined thread" computes the points coordinates on a circle,
'and the "main thread" plots the points.
'
'Principle of mutual exclusion + mutual synchronisation
'          Thread#A                XOR + <==>           Thread#B
'.....                                         .....
'MutexLock(mut)                                MutexLock(mut)
'  While Thread#A_signal <> false                While Thread#A_signal <> true
'    CondWait(cond, mut)                           CondWait(cond, mut)
'  Wend                                          Wend
'  Do_something#A_with_exclusion                 Do_something#B_with_exclusion
'  Thread#A_signal = true                        Thread#A_signal = false
'  CondSignal(cond)                              CondSignal(cond)
'MutexUnlock(mut)                              MutexUnlock(mut)
'.....                                         .....
'
'Behavior:
'- Unnecessary to pre-calculate the first point.
'- Each calculated point is plotted one time only.
'
'If you comment out the lines containing "MutexLock" and "MutexUnlock",
'"CondWait" and "CondSignal", ".ready"
'(inside "user-defined thread" or/and "main thread"),
'there will be no longer mutual exclusion nor mutual synchronization
'between computation of coordinates and plotting of points,
'and many points will not be plotted on circle (due to non coherent coordinates).

'-----------------------------------------------------------------------------------------------------

Type ThreadUDT                                   'Generic user thread UDT
	Dim handle As Any Ptr                        'Any Ptr handle to user thread
	Dim sync As Any Ptr                          'Any Ptr handle to mutex
	Dim cond As Any Ptr                          'Any Ptr handle to conditional
	Dim ready As Byte                            'Boolean to coordinates ready
	Dim quit As Byte                             'Boolean to end user thread
	Declare Static Sub Thread (ByVal As Any Ptr) 'Generic user thread procedure
	Dim procedure As Sub (ByVal As Any Ptr)      'Procedure(Any Ptr) to be executed by user thread
	Dim p As Any Ptr                             'Any Ptr to pass to procedure executed by user thread
	Const false As Byte = 0                      'Constante "false"
	Const true As Byte = Not false               'Constante "true"
End Type

Static Sub ThreadUDT.Thread (ByVal param As Any Ptr) 'Generic user thread procedure
	Dim tp As ThreadUDT Ptr = param                  'Casting to generic user thread UDT
	Do
	    Static As Integer I
	    MutexLock(tp->sync)                          'Mutex (Lock) for user thread
	    While tp->ready <> false                     'Process loop against spurious wakeups
	      CondWait(tp->cond, tp->sync)               'CondWait to receive signal from main-thread
	    Wend
	    tp->procedure(tp->p)                         'Procedure(Any Ptr) to be executed by user thread
	    I += 1
	    Locate 30, 38
	    Print I;
	    tp->ready = true                             'Set ready
	    CondSignal(tp->cond)                         'CondSignal to send signal to main thread
	    MutexUnlock(tp->sync)                        'Mutex (Unlock) for user thread
	    Sleep 5
	Loop Until tp->quit = tp->true                   'Test for ending user thread
End Sub

'-----------------------------------------------------------------------------------------------------

Type Point2D
	Dim x As Integer
	Dim y As Integer
End Type

Const x0 As Integer = 640 / 2
Const y0 As Integer = 480 / 2
Const r0 As Integer = 200
Const pi As Single = 4 * Atn(1)

Sub PointOnCircle (ByVal p As Any Ptr)
	Dim pp As Point2D Ptr = p
	Dim teta As Single = 2 * pi * Rnd
	pp->x = x0 + r0 * Cos(teta)
	Sleep 5                            'To increase possibility of uncorrelated data occurrence
	pp->y = y0 + r0 * Sin(teta)
End Sub


Screen 12
Locate 30, 2
Print "<any_key> : exit";
Locate 30, 27
Print "calculated:";
Locate 30, 54
Print "plotted:";

Dim Pptr As Point2D Ptr = New Point2D

Dim Tptr As ThreadUDT Ptr = New ThreadUDT
Tptr->sync = MutexCreate
Tptr->cond = CondCreate
Tptr->procedure = @PointOnCircle
Tptr->p = Pptr
Tptr->handle = ThreadCreate(@ThreadUDT.Thread, Tptr)

Do
	Static As Integer I
	Sleep 5
	MutexLock(Tptr->sync)              'Mutex (Lock) for main thread
	While Tptr->ready <> Tptr->true    'Process loop against spurious wakeups
	  CondWait(Tptr->cond, Tptr->sync) 'CondWait to receive signal from user-thread
	Wend
	PSet (Pptr->x, Pptr->y)            'Plotting one point
	I += 1
	Locate 30, 62
	Print I;
	Tptr->ready = Tptr->false          'Reset ready
	CondSignal(Tptr->cond)             'CondSignal to send signal to user thread
	MutexUnlock(Tptr->sync)            'Mutex (Unlock) for main thread
Loop Until Inkey <> ""
 
MutexLock(Tptr->sync)                  'Mutex (Lock) for main thread
Tptr->ready = Tptr->false              'Reset ready
Tptr->quit = Tptr->true                'Set quit
CondSignal(Tptr->cond)                 'CondSignal to send signal to user thread
MutexUnlock(Tptr->sync)                'Mutex (Unlock) for main thread

ThreadWait(Tptr->handle)
MutexDestroy(Tptr->sync)
CondDestroy(Tptr->cond)
Delete Tptr
Delete Pptr

Sleep
